package be.iminds.ilabt.jfed.testing.base;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.dropwizard.jackson.Jackson;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.lang.reflect.Method;
import java.util.*;

/**
 * JsonTestConfigEditor
 */
public class JsonTestConfigEditor<C extends ApiTestConfig> implements ApiTestConfigEditor<C> {
    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();

    @Nonnull private final Class<C> apiTestConfigClass;

    @Nonnull private Map curMap = new HashMap();

    public JsonTestConfigEditor(@Nonnull Class<C> apiTestConfigClass) {
        this.apiTestConfigClass = apiTestConfigClass;
    }

    @Nonnull
    @Override
    public List<ApiTestConfigParameter> getParameters() {
        List<ApiTestConfigParameter> res = new ArrayList<>();

        Method[] methods = apiTestConfigClass.getMethods();
        //only a very specific jackson annotation style is supported
        for (Method method : methods) {
            if (method.getName().startsWith("get")) {
                JsonProperty jsonPropertAnnotation = method.getAnnotation(JsonProperty.class);
                JsonIgnore jsonIgnoreAnnotation = method.getAnnotation(JsonIgnore.class);
                ApiTestConfigProperty propertyAnnotation = method.getAnnotation(ApiTestConfigProperty.class);
                ApiTestConfigPropertyDefault propertyDefaultAnnotation = method.getAnnotation(ApiTestConfigPropertyDefault.class);
                Nonnull nonnullAnnotation = method.getAnnotation(Nonnull.class);
                Nullable nullableAnnotation = method.getAnnotation(Nullable.class);
                boolean isProperty = jsonPropertAnnotation != null;
                if (jsonIgnoreAnnotation != null && jsonIgnoreAnnotation.value()) isProperty = false;
                if (isProperty) {
                    String name = jsonPropertAnnotation.value();
                    if (name.isEmpty() && method.getName().length() >= 4) {
                        char first = Character.toLowerCase(method.getName().charAt(3));
                        name = first + method.getName().substring(4, method.getName().length());
                    }

                    Object defaultValue = null;
                    if (propertyDefaultAnnotation != null) {
                        if (method.getReturnType().equals(String.class)) {
                            defaultValue = propertyDefaultAnnotation.value();
                        }
                        if (method.getReturnType().equals(Integer.class)) {
                            defaultValue = propertyDefaultAnnotation.longValue();
                        }
                        if (method.getReturnType().equals(Long.class)) {
                            defaultValue = propertyDefaultAnnotation.longValue();
                        }
                        if (method.getReturnType().equals(Boolean.class)) {
                            defaultValue = propertyDefaultAnnotation.booleanValue();
                        }
                        if (method.getReturnType().equals(Double.class)) {
                            defaultValue = propertyDefaultAnnotation.doubleValue();
                        }
                    }

                    ApiTestConfigParameter param = new ApiTestConfigParameter(
                            Collections.singletonList(name),
                            propertyAnnotation == null ? null : (propertyAnnotation.description().isEmpty() ? null : propertyAnnotation.description()),
                            method.getReturnType(),
                            defaultValue,
                            nonnullAnnotation != null,
                            propertyAnnotation == null ? null : (propertyAnnotation.editorHint().isEmpty() ? null : propertyAnnotation.editorHint()));
                    res.add(param);
                }
            }
        }

        return res;

//        try {
//            SchemaFactoryWrapper visitor = new SchemaFactoryWrapper();
//            MAPPER.acceptJsonFormatVisitor(apiTestConfigClass, visitor);
//            JsonSchema schema = visitor.finalSchema();
//            Map<String, JsonSchema> curShema = schema.asObjectSchema().getProperties();
//
//            List<ApiTestConfigParameter> res = new ArrayList<>();
//            for (Map.Entry<String, JsonSchema> e : curShema.entrySet()) {
//                String key = e.getKey();
//                JsonSchema subSchema = e.getValue();
//                res.add(new ApiTestConfigParameter(
//                        Collections.singletonList(key),
//                        "TODO",
//                        subSchema.getType()+"",
//                        "TODO",
//                        false,
//                        "TODO"
//                ));
//            }
//            return res;
//        } catch (JsonMappingException e) {
//            throw new RuntimeException("Failed to process config class "+apiTestConfigClass.getName(), e);
//        }
    }

    @Override
    public void setParameter(@Nonnull ApiTestConfigParameter parameter, @Nullable Object value) {
        Map m = getSubMap(parameter);
        m.put(parameter.getLastName(), value);
    }

    @Nullable
    @Override
    public Object getParameter(@Nonnull ApiTestConfigParameter parameter) {
        Map m = getSubMap(parameter);
        return m.get(parameter.getLastName());
    }

    /** will add create sub maps as needed */
    @Nonnull
    private Map getSubMap(@Nonnull ApiTestConfigParameter parameter) {
        List<String> keys = parameter.getName();
        Map res = curMap;
        for (int i = 0; i < keys.size()-1; i++) {
            String key = keys.get(i);
            Object val = res.get(key);
            if (val == null) {
                Map newMap = new HashMap();
                res.put(key, newMap);
                res = newMap;
            } else {
                if (val instanceof Map) {
                    res = (Map) val;
                } else {
                    throw new RuntimeException("Invalid nesting! Something must be wrong with the parameters. keys="+keys+" i="+i+" val="+val+" valclass="+val.getClass().getName());
                }
            }
        }
        return res;
    }

    @Override
    public void setConfig(@Nonnull C apiTestConfig) {
        curMap = MAPPER.convertValue(apiTestConfig, Map.class);
    }

    @Nonnull
    @Override
    public C build() {
        return MAPPER.convertValue(curMap, apiTestConfigClass);
    }
}
